/***************************************************************************
                          dfg1394.h  -  description
                             -------------------
 Copyright (c) 2003 Norbert Wagner <nw@softwarekombinat.de>



    created              : 2003-03-12 Norbert Wagner <nw@softwarekombinat.de>
    modified             : 2003-10-07 Arne Caspari
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef __DFG1394_H__
#define __DFG1394_H__

#include <linux/videodev.h>
#include <linux/types.h>
#include <linux/device.h>


/* forward declarations */
struct hpsb_iso;
struct hpsb_host;


/*
 * =========================================================================
 *
 * DEFINES
 *
 * =========================================================================
 */

/*******************************************************/
/** @name Module Defines                              */
/*  @{                                                */

/** @brief The module's name, used in various places */
#define MODULE_NAME "dfg1394-1"

/** @brief Max. number of buffers a client may use to render */
#define DFG1394_STREAM_MAX_BUFFERS 16

/** @brief The converter's unique bus ID */
#define DFG1394_NODE_UID 0x0007480000526F6E//0x0007480044281067//

/** @brief Checks if the specified expression was a success */
#define IS_SUCCESS(expr) ((expr) >= 0)

/** @brief Checks if the specified expression was a failure */
#define IS_FAILURE(expr) ((expr) < 0)

/** @brief Return code to indicate a successfull completion */
#define SUCCESS 0

/*@}*/

/*******************************************************/
/** @name printk Helpers                              */
/*  @{                                                */

/** @brief Enable debug output */
#define LVL_DEBUG 1


#ifdef LVL_DEBUG
#define PRINT_DEBUG(fmt,arg...) \
        printk(KERN_DEBUG "[" MODULE_NAME "] " fmt, ##arg)
#else
#define PRINT_DEBUG(fmt,arg...)
#endif

#define PRINT_ERR(fmt,arg...) \
        printk(KERN_ERR "[" MODULE_NAME "] " fmt, ##arg)

#define PRINT_WARNING(fmt,arg...) \
        printk(KERN_WARNING "[" MODULE_NAME "] " fmt, ##arg)

#define PRINT_NOTICE(fmt,arg...) \
        printk(KERN_NOTICE "[" MODULE_NAME "] " fmt, ##arg)

#define PRINT_INFO(fmt,arg...) \
        printk(KERN_INFO "[" MODULE_NAME "] " fmt, ##arg)

#define PRINT_TRACE(fmt,arg...)  \
        printk(KERN_INFO "[" MODULE_NAME  "." __FILE__ ":%d]: " fmt, __LINE__, ##arg)

#define PRINT_TRACE1(fmt,arg...) \
        printk(KERN_INFO "[" MODULE_NAME  "." __FILE__ ":%d]: " fmt, __LINE__ )

/*@}*/


/** boolean type */
typedef enum { TRUE = 1, FALSE = 0 } dfg1394_bool_t;


/**
 * @brief Buffer to capture data into
 */
struct dfg1394_buffer
{
	/** @brief video for linux buffer */
	struct v4l2_buffer v4l2buf;

	/**  @brief TRUE if buffer was requested by VIDEOC_REQBUFS call*/
	dfg1394_bool_t is_requested;
	
	/**  @brief virtual address of buffer data */
	u8*  data;	

	/**  @brief the size of the frame to be captured into this buffer */
	u32  frame_size;

	/**  @brief the size of one line */
	u32  line_size;

	/**  @brief number of vma users */
	int vma_user_count;

	/* ================================================== 
	 *  capture state
         */

	/**  @brief the position the next byte is written to */ 
	unsigned int   offset;

	/**  @brief list management */
	struct list_head list_entry;
};

typedef struct dfg1394_buffer dfg1394_buffer_t;



/***********************************************************************\
 *                                                                     *
 *  dfg1394_node                                                       *
 *                                                                     *
 \**********************************************************************/

/*******************************************************/
/** @name dfg1394_node related types                   */
/*  @{                                                 */

/**
 * Frequencies used by the current input channel
 */
typedef enum {
	DFG1394_FREQ_PAL_50  = 0,
	DFG1394_FREQ_NTSC_60 = 1,
	DFG1394_FREQ__COUNT,
	DFG1394_FREQ_DETECT  = 2,
	DFG1394_FREQ_DEFAULT = DFG1394_FREQ_PAL_50
} dfg1394_freq_t;

/**
 * Hardware videomode of the DFG1394-1 Video Converter
 */
typedef enum {
	DFG1394_PIX_MODE__FIRST      = 0,
	DFG1394_PIX_MODE_320x240_400 = DFG1394_PIX_MODE__FIRST,
	DFG1394_PIX_MODE_320x240_411 = 1,
	DFG1394_PIX_MODE_320x240_422 = 2,
	DFG1394_PIX_MODE_640x480_400 = 3,
	DFG1394_PIX_MODE_640x480_411 = 4,
	DFG1394_PIX_MODE_640x480_422 = 5,
	DFG1394_PIX_MODE_768x576_400 = 6,
	DFG1394_PIX_MODE_768x576_411 = 7,
	DFG1394_PIX_MODE_768x576_422 = 8,
	DFG1394_PIX_MODE__COUNT,
	DFG1394_PIX_MODE_DEFAULT     = DFG1394_PIX_MODE_640x480_422 
} dfg1394_pix_mode_t;

/**
 * Input channels of the DFG1394-1 Video Converter
 */
typedef enum {
	DFG1394_INPUT_COMPOSITE_0 = 0x00,
	DFG1394_INPUT_COMPOSITE_1 = 0x01,
	DFG1394_INPUT_COMPOSITE_2 = 0x02,
	DFG1394_INPUT_COMPOSITE_3 = 0x03,
	DFG1394_INPUT_COMPOSITE_4 = 0x04,
	DFG1394_INPUT_COMPOSITE_5 = 0x05,
	DFG1394_INPUT_SVIDEO_0    = 0x06,
	DFG1394_INPUT_SVIDEO_1    = 0x07,
	DFG1394_INPUT_SVIDEO_2    = 0x08,
	DFG1394_INPUT_SVIDEO_3    = 0x09,
	DFG1394_INPUT__COUNT,
	DFG1394_INPUT_DETECT      = 0xff,
	DFG1394_INPUT_DEFAULT     = DFG1394_INPUT_SVIDEO_3
//DFG1394_INPUT_COMPOSITE_0// 
} dfg1394_input_t;

/**
 * Field type
 */
typedef enum {
	DFG1394_FT_UNKNOWN = 0,
	DFG1394_FT_ODD     = 1,
	DFG1394_FT_EVEN    = 2
} dfg1394_field_type_t;


typedef u64 dfg1394_uid_t;



/**
 * Node Flags
 */
enum {
	/** set if the node is transmitting */
	DFG1394_NODE_FLAG__TRANSMITTING   = 0,

	/** set when a pending fcp request was completed */
	DFG1394_NODE_FLAG__FCP_DONE       = 1
};

struct _dfg1394_firmware_version
{
	unsigned char reserved1;
	unsigned char reserved2;
	unsigned char major;
	unsigned char minor;
};

typedef struct _dfg1394_firmware_version dfg1394_firmware_version_t;




/*@}*/

/**
 * @brief Device representation of the Node and Firewire Layers
 */
struct dfg1394_node 
{	
	/*  ================================================== */
	/** @name      IEEE 1394 Bus Address                   */
	/*  @{                                                 */

	/** the host this node is connected to */
	struct hpsb_host *host;  

        /** the host generation (used to detect bus resets ) */
	u32    generation;       
                                 
	/** the nodeId _without_ busId */
	int    node;             

        /*@}*/

	/*  ================================================== */
	/** @name      DFG/1394-1 Device Configuration         */
	/*  @{                                                 */

	/** iso channel the device is broadcasting at */
	u8                 iso_channel;

	/** configured input channel */
	dfg1394_input_t    input;

	/** configured pixel mode */
	dfg1394_pix_mode_t pix_mode;

	/** configured frequency */
	dfg1394_freq_t     frequency;

	dfg1394_firmware_version_t firmware_version;
		

	/** used to synchronize concurrent access from
	   different processes on SMP systems 
	 */
	spinlock_t lock;

        /*@}*/

	/*  ================================================== */
	/** @name      Communication related                   */
	/*  @{                                                 */

	/** contains the fcp_response if we have any */
	u64 fcp_response;                    
                                             
	/** to wait for FCP response */
	wait_queue_head_t fcp_response_wait; 

	/** protects fcp_response */
	struct semaphore fcp_busy_sema;      

	/** ieee1394 highlevel iso context */
	struct hpsb_iso* iso; 

        /** misc flags */
	unsigned long flags;

        /*@}*/

	/*  ================================================== */
	/** @name      Current stream state                    

	    @warning These members are accessed from 
                     IRQ context!                                          
                                                               */
	/*  @{                                                 */
	
	/** TRUE if the node is synchronizing */
	dfg1394_bool_t       stream_is_syncing ;

	/** the current field type */
	dfg1394_field_type_t stream_field_type ;
	
	/** the number of frames seen in the current transmission */
	unsigned long        stream_frame_count;

	/** last sy value */
	u8                   stream_last_sy;

	/** number of bytes left in the current line */
	int                  stream_line_left;

	/** the buffer we are currently streaming in */
	dfg1394_buffer_t*    stream_buffer;

	/** used to synchronize access to data accessed
	    by interrupt handlers
	*/ 
	spinlock_t stream_lock;

        /*@}*/

	/*  ================================================== */
	/** @name      Statistics                              */
	/*  @{                                                 */
	
	/** number of frames captured */
	atomic_t        stat_frames_captured;

	/** number of lost frames */
        atomic_t        stat_frames_lost;

	/** number of dropped frames */
	atomic_t        stat_frames_dropped;

};

typedef struct dfg1394_node dfg1394_node_t;


/*******************************************************/
/** @name dfg1394_node Methods                         */
/*  @{                                                 */

extern void dfg1394_node__initialize( dfg1394_node_t* node );

extern int dfg1394_node__start( dfg1394_node_t* node );

extern void dfg1394_node__drop_dead( dfg1394_node_t* node );

extern int dfg1394_node__set_frequency( 
	dfg1394_node_t* node, dfg1394_freq_t req_freq );

extern int dfg1394_node__set_brightness( 
	dfg1394_node_t *node, u8 value );

extern int dfg1394_node__get_brightness(
	dfg1394_node_t *node, u8 *value );

extern int dfg1394_node__set_contrast(
	dfg1394_node_t *node, u8 value );

extern int dfg1394_node__get_contrast(
	dfg1394_node_t *node, u8 *value );

extern int dfg1394_node__set_hue(
	dfg1394_node_t *node, u8 value );

extern int dfg1394_node__get_hue(
	dfg1394_node_t *node, u8 *value );

extern int dfg1394_node__set_saturation(
	dfg1394_node_t *node, u8 value );

extern int dfg1394_node__get_saturation(
	dfg1394_node_t *node, u8 *value );

extern int dfg1394_node__get_firmware_version( dfg1394_node_t *node );

extern dfg1394_freq_t dfg1394_node__get_frequency(
	dfg1394_node_t* node);

extern int dfg1394_node__set_input( 
	dfg1394_node_t* node, dfg1394_input_t req_input );

extern dfg1394_input_t dfg1394_node__get_input(
	dfg1394_node_t* node);

extern int dfg1394_node__set_mode( 
	dfg1394_node_t* node, dfg1394_pix_mode_t mode );

extern dfg1394_pix_mode_t dfg1394_node__get_mode(dfg1394_node_t* node);

extern int dfg1394_node__start_transmission( 
	dfg1394_node_t* node, u8 channel );

extern int dfg1394_node__stop_transmission( 
	dfg1394_node_t* node );

/**
 * @brief Checks if the converter is transmitting
 */
#define dfg1394_node__is_transmitting(node) \
	test_bit( DFG1394_NODE_FLAG__TRANSMITTING, &(node)->flags )


/**
 * @brief Gets the number of captured frames
 */
#define dfg1394_node__get_frames_captured(node) \
        atomic_read( &(node)->stat_frames_captured )

/**
 * @brief  Gets the number of lost frames
 */
#define dfg1394_node__get_frames_lost(node) \
        atomic_read( &(node)->stat_frames_lost )

/**
 * @brief  Gets the number of dropped frames
 */
#define dfg1394_node__get_frames_dropped(node) \
        atomic_read( &(node)->stat_frames_dropped )


/* ------------------------------------
 * Callbacks used by the firewire layer:
 */

extern void dfg1394_node__process_fcp_response( 
	dfg1394_node_t *node, u64 response );

extern void  dfg1394_node__process_iso_packet( 
	dfg1394_node_t *node, u8 sy, u8* data, u32 length );


/*@}*/


/***********************************************************************\
 *                                                                     *
 *  dfg1394_device                                                     *
 *                                                                     *
 \**********************************************************************/

/*******************************************************/
/** @name dfg1394_device related types                 */
/*  @{                                                 */

/**
 * @brief Device Flags
 */
enum {
	/** Set if the device is a ghost */
	DFG1394_DEVICE_FLAG__GHOST = 0
};


/*@}*/

/**
 *	@brief The DFG1394-1 device structure.
 *
 *      There's one struct for each physically connected Video Converter.
 */
struct dfg1394_device
{
	/** 
	 * @brief Lowlevel node data
	 *
	 * This must be first member!
	 */
	dfg1394_node_t node;


	/*  ================================================== */
	/** @name      General Linux Device related            */
	/*  @{                                                 */

	/** @brief Connection into Linux Driver Architecture:  */
	struct device* device;

	/** @brief the number of clients on this device */
	u32 client_count;

	/* @brief contains the GHOST flag */
	unsigned long flags;

	/** @brief used to synchronize concurrent access from
	   different processes on SMP systems 
	 */
	spinlock_t lock;

	/** @brief used to serialize system calls */
	struct semaphore sema;      

	/** Hook into dfg1394_driver device list */
	struct list_head list_entry;

        /*@}*/

	/*  ================================================== */
	/** @name      V4L2 Device related                     */
	/*  @{                                                 */

	/** @brief the video for linux device object */
        struct video_device       vdev; 

	/** @brief the current v4l2 pixel format */
	struct v4l2_pix_format    pix_format;

        /*@}*/

	/*  ================================================== */
	/** @name      Capture buffers and capture state       */
	/*  @{                                                 */

	/* the client with the right to capture (there can be only one) */
	struct dfg1394_client* cap_client;

	/** the capture buffer instances */
	dfg1394_buffer_t cap_buffers[DFG1394_STREAM_MAX_BUFFERS];

	/** empty buffers queued for capturing  */
	struct list_head cap_free_queue;

	/** full buffers queued for userspace  */
	struct list_head cap_done_queue;

	/** to wait for done buffers */
	wait_queue_head_t cap_done_wait;

	/** to protect the capture state - it's accessed from 
	    our irq handler! */
	spinlock_t cap_lock;

        /*@}*/
};

typedef struct dfg1394_device dfg1394_device_t;


/* forward declarations */
struct dfg1394_client;

/*******************************************************/
/** @name dfg1394_device Methods                       */
/*  @{                                                 */


extern dfg1394_device_t* dfg1394_device__create( struct device* device );

extern int dfg1394_device__destroy(dfg1394_device_t* dev);

extern int dfg1394_device__register( dfg1394_device_t* dev, int nr );

extern int dfg1394_device__unregister( dfg1394_device_t* dev );

/**
 * @brief Checks if the device is only a ghost
 */
#define dfg1394_device__is_ghost(dev) \
	test_bit( DFG1394_DEVICE_FLAG__GHOST, &(dev)->flags )

extern int dfg1394_device__ghostify( dfg1394_device_t* dev );

extern int dfg1394_device__resurrect(
	dfg1394_device_t* dev, struct device* device  );


extern dfg1394_buffer_t* dfg1394_device__pop_empty_buffer(
	dfg1394_device_t* dev);

extern void dfg1394_device__push_filled_buffer( 
	dfg1394_device_t* dev,  
	dfg1394_buffer_t* buf, u32 frame_count);

/*@}*/

/***********************************************************************\
 *                                                                     *
 *  dfg1394_client                                                     *
 *                                                                     *
 \**********************************************************************/


/**
 * @brief Holds data associated with a driver's client.
 *
 * There's one struct for each open connection
 * to a userspace programm.
 */
struct dfg1394_client
{
	/** @brief Flag: Is the client opened? */
	dfg1394_bool_t is_open      : 1;

	/** @brief Flag: Is the client blocking? */
	dfg1394_bool_t is_blocking  : 1;

	/** @brief The opened device object */
	dfg1394_device_t* dev;

};
typedef struct dfg1394_client dfg1394_client_t;

/*******************************************************/
/** @name dfg1394_client Methods                       */
/*  @{                                                 */


/**
 * @brief Creates a new dfg1394_client structure and initializes its members 
 */
dfg1394_client_t* dfg1394_client__create( 
	dfg1394_device_t* dev, dfg1394_bool_t is_blocking );

/**
 * @brief Destroys an existing dfg1394_client structure
 */
void dfg1394_client__destroy( dfg1394_client_t* client );

/*@}*/


/***********************************************************************\
 *                                                                     *
 *  dfg1394_driver                                                     *
 *                                                                     *
 \**********************************************************************/


/**
 * @brief Holds the driver's global data
 */
struct dfg1394_driver
{
	/** @brief the IEEE 1394 Driver device_driver object */
	struct hpsb_protocol_driver* driver;

	/** @brief List of device objects */
	struct list_head devices;

	/** @brief Number of associated devices */
	u32 device_count;
};

typedef struct dfg1394_driver dfg1394_driver_t; 

/*******************************************************/
/** @name dfg1394_driver Methods                       */
/*  @{                                                 */


extern int dfg1394_driver__create( void );

extern int dfg1394_driver__destroy( void );

/*@}*/




/***********************************************************************\
 *                                                                     *
 *  firewire layer                                                     *
 *                                                                     *
 \**********************************************************************/

/*******************************************************/
/** @name The Firewire Layer                           */
/*  @{                                                 */


extern int dfg1394_fw_start_iso_receive( struct dfg1394_node* node  );


extern int dfg1394_fw_stop_iso_receive( struct dfg1394_node* node  );

extern int dfg1394_fw_send_fcp_command( 
	struct dfg1394_node* node, u8 cmd, u8 arg1, u8 arg2 );

extern int dfg1394_fw_scan_for_device( struct dfg1394_node* node );   


extern int dfg1394_fw_init(void);

extern int dfg1394_fw_exit(void);

extern int dfg1394_fw_send_fcp3_command( 
	dfg1394_node_t* node,	u8 cmd, u8 arg1, u8 arg2, u8 arg3 );


/*@}*/

#endif
